@@ -2,17 +2,20 @@
 #include <structmember.h>
 
 /*
-Persistent/Immutable data structures. Unfortunately I have not been able to come
-up with an implementation that is 100% immutable due to the ref counts used by
-Python internally so the GIL must still be at work...
+Persistent/Immutable/Functional vector and helper types. 
 
-To the end user they should appear immutable at least.
+Please note that they are anything but immutable at this level since
+there is a whole lot of reference counting going on. That's the way
+CPython works though and the GIL makes them appear immutable.
+
+To the programmer using them from Python they appear immutable and
+behave immutably at least.
 
 Naming conventions
 ------------------
 initpyrsistentc - This is the method that initializes the whole module
 pyrsistent_* -    Methods part of the interface
-Pvector_*    -    Instance methods of the PVector object
+<typename>_* -    Instance methods of types. For examle PVector_append(...)
 
 All other methods are camel cased without prefix. All methods are static, none should
 require to be exposed outside of this module. 
@@ -22,6 +25,7 @@ require to be exposed outside of this module.
 #define BIT_MASK (BRANCH_FACTOR - 1)
 
 static PyTypeObject PVectorType;
+static PyTypeObject PVectorEvolverType;
 
 int SHIFT = 0;
 
@@ -47,6 +51,14 @@ typedef struct {
   VNode *tail;
 } PVector;
 
+typedef struct {
+  PyObject_HEAD
+  PVector* originalVector;
+  PVector* newVector;
+  PyObject* appendList;
+} PVectorEvolver;
+
+
 static PVector* EMPTY_VECTOR = NULL;
 static PyObject* pmap_factory_fn = NULL;
 static PyObject* set_in_fn_name = NULL;
@@ -77,6 +89,11 @@ static PyMemberDef PVector_members[] = {
 #define debug(...)
 // #define debug printf
 
+#define NODE_REF_COUNT(n) ((n)->refCount)
+#define SET_NODE_REF_COUNT(n, c) (NODE_REF_COUNT(n) = (c))
+#define INC_NODE_REF_COUNT(n) (NODE_REF_COUNT(n)++)
+#define DEC_NODE_REF_COUNT(n) (NODE_REF_COUNT(n)--)
+
 static VNode* allocNode(void) {
   if(nodeCache.size > 0) {
     nodeCache.size--;
@@ -98,7 +115,7 @@ static void freeNode(VNode *node) {
 static VNode* newNode(void) {
   VNode* result = allocNode();
   memset(result, 0x0, sizeof(VNode));
-  result->refCount = 1;
+  SET_NODE_REF_COUNT(result, 1);
   debug("newNode() %p\n", result);
   return result;
 }
@@ -113,17 +130,20 @@ static VNode* copyNode(VNode* source) {
   
   for(i = 0; i < BRANCH_FACTOR; i++) {
     if(result->items[i] != NULL) {
-      ((VNode*)result->items[i])->refCount++;
+      INC_NODE_REF_COUNT((VNode*)result->items[i]);
     }
   }
 
-  result->refCount = 1;
+  SET_NODE_REF_COUNT(result, 1);
   return result;
 }
 
 static PVector* emptyNewPvec(void);
 static PVector* copyPVector(PVector *original);
-static void extend_with_item(PVector *newVec, PyObject *item);
+static void extendWithItem(PVector *newVec, PyObject *item);
+
+static PyObject *PVectorEvolver_pvector(PVectorEvolver *);
+static int PVectorEvolver_set_item(PVectorEvolver *, PyObject*, PyObject*);
 
 static Py_ssize_t PVector_len(PVector *self) {
   return self->count;
@@ -156,12 +176,11 @@ static VNode* nodeFor(PVector *self, int i){
 
 static PyObject* _get_item(PVector *self, Py_ssize_t pos) {
   VNode* node = nodeFor((PVector*)self, pos);
-
+  PyObject *result = NULL;
   if(node != NULL) {
-    return node->items[pos & BIT_MASK];
+    result = node->items[pos & BIT_MASK];
   }
-
-  return NULL;
+  return result;
 }
 
 /*
@@ -182,12 +201,13 @@ static void releaseNode(int level, VNode *node) {
     return;
   }
 
-  debug("releaseNode(): node=%p, level=%i, refCount=%i\n", node, level, node->refCount);
+  debug("releaseNode(): node=%p, level=%i, refCount=%i\n", node, level, NODE_REF_COUNT(node));
 
   int i;
 
-  node->refCount--;
-  if(node->refCount == 0) {
+  DEC_NODE_REF_COUNT(node);
+  debug("Refcount when trying to release: %u\n", NODE_REF_COUNT(node));
+  if(NODE_REF_COUNT(node) == 0) {
     if(level > 0) {
       for(i = 0; i < BRANCH_FACTOR; i++) {
         if(node->items[i] != NULL) {
@@ -212,20 +232,20 @@ static void releaseNode(int level, VNode *node) {
  those if needed.
 */
 static void PVector_dealloc(PVector *self) {
-  debug("Dealloc(): self=%p, self->count=%i, tail->refCount=%i, root->refCount=%i, self->shift=%i, self->tail=%p, self->root=%p\n",
-   self, self->count, self->tail->refCount, self->root->refCount, self->shift, self->tail, self->root, self->root);
+  debug("Dealloc(): self=%p, self->count=%u, tail->refCount=%u, root->refCount=%u, self->shift=%u, self->tail=%p, self->root=%p\n",
+        self, self->count, NODE_REF_COUNT(self->tail), NODE_REF_COUNT(self->root), self->shift, self->tail, self->root);
 
   PyObject_GC_UnTrack((PyObject*)self);
-  Py_TRASHCAN_SAFE_BEGIN(self)
+  Py_TRASHCAN_SAFE_BEGIN(self);
 
   releaseNode(0, self->tail);
   releaseNode(self->shift, self->root);
   
   PyObject_GC_Del(self);
-  Py_TRASHCAN_SAFE_END(self)
+  Py_TRASHCAN_SAFE_END(self);
 }
 
-static PyObject * to_list(PVector *self) {
+static PyObject *toList(PVector *self) {
   Py_ssize_t i;
   PyObject *list = PyList_New(self->count);
   for (i = 0; i < self->count; ++i) {
@@ -239,7 +259,7 @@ static PyObject * to_list(PVector *self) {
 
 static PyObject *PVector_repr(PVector *self) {
   // Reuse the list repr code, a bit less efficient but saves some code
-  PyObject *list = to_list(self);
+  PyObject *list = toList(self);
   PyObject *list_repr = PyObject_Repr(list);
   Py_DECREF(list);
 
@@ -378,7 +398,7 @@ static PyObject* PVector_repeat(PVector *self, Py_ssize_t n) {
     PVector *newVec = copyPVector(self);
     for(i=0; i<(n-1); i++) {
       for(j=0; j<self->count; j++) {
-        extend_with_item(newVec, PVector_get_item(self, j));
+        extendWithItem(newVec, PVector_get_item(self, j));
       }
     }
     return (PyObject*)newVec;
@@ -464,7 +484,7 @@ static PyObject* PVector_pickle_reduce(PVector *self) {
   PyObject* pvector_fn = PyObject_GetAttrString(module, "pvector");
   Py_DECREF(module);
 
-  PyObject *list = to_list(self);
+  PyObject *list = toList(self);
   PyObject *arg_tuple = PyTuple_New(1);
   PyTuple_SET_ITEM(arg_tuple, 0, list);
 
@@ -475,6 +495,42 @@ static PyObject* PVector_pickle_reduce(PVector *self) {
   return result_tuple;
 }
 
+static PVector* rawCopyPVector(PVector* vector) {
+  PVector* newVector = PyObject_GC_New(PVector, &PVectorType);
+  newVector->count = vector->count;
+  newVector->shift = vector->shift;
+  newVector->root = vector->root;
+  newVector->tail = vector->tail;
+  PyObject_GC_Track((PyObject*)newVector);
+  return newVector;
+}
+
+static void initializeEvolver(PVectorEvolver* evolver, PVector* vector, PyObject* appendList) {
+  // Need to hold a reference to the underlying vector to manage
+  // the ref counting properly.
+  evolver->originalVector = vector;
+  evolver->newVector = vector;
+
+  if(appendList == NULL) {
+    evolver->appendList = PyList_New(0);
+  } else {
+    evolver->appendList = appendList;
+  }
+}
+
+static PyObject * PVector_evolver(PVector *self) {
+  PVectorEvolver *evolver = PyObject_GC_New(PVectorEvolver, &PVectorEvolverType);
+  if (evolver == NULL) {
+    return NULL;
+  }
+  
+  PyObject_GC_Track(evolver);
+  initializeEvolver(evolver, self, NULL);
+  Py_INCREF(self);
+  return (PyObject *)evolver;
+}
+
+
 static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
   memcpy(dest, src, BRANCH_FACTOR * sizeof(void*));
   dest[pos] = obj;
@@ -486,6 +542,8 @@ static PyObject* PVector_set_in(PVector *self, PyObject *obj);
 
 static PyObject* PVector_set(PVector *self, PyObject *obj);
 
+static PyObject* PVector_mset(PVector *self, PyObject *args);
+
 static PyObject* PVector_subscript(PVector* self, PyObject* item);
 
 static PyObject* PVector_extend(PVector *self, PyObject *args);
@@ -495,6 +553,7 @@ static PySequenceMethods PVector_sequence_methods = {
     (binaryfunc)PVector_extend,      /* sq_concat */
     (ssizeargfunc)PVector_repeat,    /* sq_repeat */
     (ssizeargfunc)PVector_get_item,  /* sq_item */
+    // TODO might want to move the slice function to here
     NULL,                            /* sq_slice */
     NULL,                            /* sq_ass_item */
     NULL,                            /* sq_ass_slice */
@@ -513,12 +572,13 @@ static PyMappingMethods PVector_mapping_methods = {
 static PyMethodDef PVector_methods[] = {
 	{"append",      (PyCFunction)PVector_append, METH_O,       "Appends an element"},
 	{"set",         (PyCFunction)PVector_set, METH_VARARGS, "Inserts an element at the specified position"},
-	{"__getitem__", (PyCFunction)PVector_subscript, METH_O|METH_COEXIST, "Subscript"},
 	{"extend",      (PyCFunction)PVector_extend, METH_O|METH_COEXIST, "Extend"},
 	{"set_in",      (PyCFunction)PVector_set_in, METH_VARARGS, "Insert an element in a nested structure"},
 	{"index",       (PyCFunction)PVector_index, METH_VARARGS, "Return first index of value"},
 	{"count",       (PyCFunction)PVector_count, METH_O, "Return number of occurrences of value"},
         {"__reduce__",  (PyCFunction)PVector_pickle_reduce, METH_NOARGS, "Pickle support method"},
+        {"evolver",     (PyCFunction)PVector_evolver, METH_NOARGS, "Return new evolver for pvector"},
+	{"mset",        (PyCFunction)PVector_mset, METH_VARARGS, "Inserts multiple elements at the specified positions"},
 	{NULL}
 };
 
@@ -597,6 +657,7 @@ static void incRefs(PyObject **obj) {
   }
 }
 
+
 static PVector* newPvec(unsigned int count, unsigned int shift, VNode *root) {
   PVector *pvec = PyObject_GC_New(PVector, &PVectorType);
   debug("pymem alloc_copy %x, ref cnt: %u\n", pvec, pvec->ob_refcnt);
@@ -610,7 +671,7 @@ static PVector* newPvec(unsigned int count, unsigned int shift, VNode *root) {
 
 static VNode* newPath(unsigned int level, VNode* node){
   if(level == 0) {
-    node->refCount++;
+    INC_NODE_REF_COUNT(node);
     return node;
   }
   
@@ -620,47 +681,47 @@ static VNode* newPath(unsigned int level, VNode* node){
 }
 
 static VNode* pushTail(unsigned int level, unsigned int count, VNode* parent, VNode* tail) {
-  int sub_index = ((count - 1) >> level) & BIT_MASK;
+  int subIndex = ((count - 1) >> level) & BIT_MASK;
   VNode* result = copyNode(parent);
   VNode* nodeToInsert;
   VNode* child;
-  debug("pushTail(): count = %i, sub_index = %i\n", count, sub_index);
+  debug("pushTail(): count = %i, subIndex = %i\n", count, subIndex);
 
   if(level == SHIFT) {
     // We're at the bottom
-    tail->refCount++;
+    INC_NODE_REF_COUNT(tail);
     nodeToInsert = tail;
   } else {
     // More levels available in the tree
-    child = parent->items[sub_index];
+    child = parent->items[subIndex];
 
     if(child != NULL) {
       nodeToInsert = pushTail(level - SHIFT, count, child, tail);
 
-      // Need to make an adjustment of the refCount for the child node here since
+      // Need to make an adjustment of the ref COUNT for the child node here since
       // it was incremented in an earlier stage when the node was copied. Now the child
       // node will be part of the path copy so the number of references to the original
       // child will not increase at all.
-      child->refCount--;
+      DEC_NODE_REF_COUNT(child);
     } else {
       nodeToInsert = newPath(level - SHIFT, tail);
     }
   }
   
-  result->items[sub_index] = nodeToInsert;
+  result->items[subIndex] = nodeToInsert;
   return result;
 }
 
 static PVector* copyPVector(PVector *original) {
   PVector *newVec = newPvec(original->count, original->shift, original->root);
-  original->root->refCount++;
+  INC_NODE_REF_COUNT(original->root);
   memcpy(newVec->tail->items, original->tail->items, TAIL_SIZE(original) * sizeof(void*));
   incRefs((PyObject**)newVec->tail->items);
   return newVec;
 }
 
 /* Does not steal a reference, this must be managed outside of this function */
-static void extend_with_item(PVector *newVec, PyObject *item) {
+static void extendWithItem(PVector *newVec, PyObject *item) {
   unsigned int tail_size = TAIL_SIZE(newVec);
 
   if(tail_size >= BRANCH_FACTOR) {
@@ -677,9 +738,9 @@ static void extend_with_item(PVector *newVec, PyObject *item) {
 
     newVec->root = new_root;
 
-    // Need to adjust the refCount of the old tail here since no new references were
+    // Need to adjust the ref count of the old tail here since no new references were
     // actually created, we just moved the tail.
-    newVec->tail->refCount--;
+    DEC_NODE_REF_COUNT(newVec->tail);
     newVec->tail = newNode();
     tail_size = 0;
   }
@@ -697,40 +758,40 @@ static void extend_with_item(PVector *newVec, PyObject *item) {
 #endif
 
 static PyObject *PVector_subscript(PVector* self, PyObject* item) {
-    if (PyIndex_Check(item)) {
-      Py_ssize_t i = PyNumber_AsSsize_t(item, PyExc_IndexError);
-      if (i == -1 && PyErr_Occurred()) {
-         return NULL;
-      }
-
-      return PVector_get_item(self, i);
-    } else if (PySlice_Check(item)) {
-      Py_ssize_t start, stop, step, slicelength, cur, i;
-      if (PySlice_GetIndicesEx(SLICE_CAST item, self->count,
-             &start, &stop, &step, &slicelength) < 0) {
-         return NULL;
-      }
-
-      debug("start=%i, stop=%i, step=%i\n", start, stop, step);
-
-      if (slicelength <= 0) {
-        Py_INCREF(EMPTY_VECTOR);
-        return (PyObject*)EMPTY_VECTOR;
-      } else if((slicelength == self->count) && (step > 0)) {
-        Py_INCREF(self);
-        return (PyObject*)self;
-      } else {
-        PVector *newVec = copyPVector(EMPTY_VECTOR);
-        for (cur=start, i=0; i<slicelength; cur += (size_t)step, i++) {
-          extend_with_item(newVec, PVector_get_item(self, cur));
-        }
-
-        return (PyObject*)newVec;
-      }
-    } else {
-      PyErr_Format(PyExc_TypeError, "pvector indices must be integers, not %.200s", Py_TYPE(item)->tp_name);
+  if (PyIndex_Check(item)) {
+    Py_ssize_t i = PyNumber_AsSsize_t(item, PyExc_IndexError);
+    if (i == -1 && PyErr_Occurred()) {
       return NULL;
     }
+    
+    return PVector_get_item(self, i);
+  } else if (PySlice_Check(item)) {
+    Py_ssize_t start, stop, step, slicelength, cur, i;
+    if (PySlice_GetIndicesEx(SLICE_CAST item, self->count,
+                             &start, &stop, &step, &slicelength) < 0) {
+      return NULL;
+    }
+    
+    debug("start=%i, stop=%i, step=%i\n", start, stop, step);
+    
+    if (slicelength <= 0) {
+      Py_INCREF(EMPTY_VECTOR);
+      return (PyObject*)EMPTY_VECTOR;
+    } else if((slicelength == self->count) && (step > 0)) {
+      Py_INCREF(self);
+      return (PyObject*)self;
+    } else {
+      PVector *newVec = copyPVector(EMPTY_VECTOR);
+      for (cur=start, i=0; i<slicelength; cur += (size_t)step, i++) {
+        extendWithItem(newVec, PVector_get_item(self, cur));
+      }
+      
+      return (PyObject*)newVec;
+    }
+  } else {
+    PyErr_Format(PyExc_TypeError, "pvector indices must be integers, not %.200s", Py_TYPE(item)->tp_name);
+    return NULL;
+  }
 } 
 
 /* A hack to get some of the error handling code away from the function
@@ -775,7 +836,7 @@ static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
     } else {
       PVector *newVec = copyPVector(self);
       while(item != NULL) {
-        extend_with_item(newVec, item);
+        extendWithItem(newVec, item);
         item = iternext(it);
       }
 
@@ -797,7 +858,7 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
   // Does the new object fit in the tail? If so, take a copy of the tail and
   // insert the new element in that.
   if(tail_size < BRANCH_FACTOR) {
-    self->root->refCount++;
+    INC_NODE_REF_COUNT(self->root);
     PVector *new_pvec = newPvec(self->count + 1, self->shift, self->root);
     copyInsert(new_pvec->tail->items, self->tail->items, tail_size, obj);
     incRefs((PyObject**)new_pvec->tail->items);
@@ -813,7 +874,7 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
   if(ROOT_NODE_FULL(self)) {
     new_root = newNode();
     new_root->items[0] = self->root;
-    self->root->refCount++;
+    INC_NODE_REF_COUNT(self->root);
     new_root->items[1] = newPath(self->shift, self->tail);
     new_shift = self->shift + SHIFT;
   } else {
@@ -829,24 +890,24 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
 }
 
 static VNode* doSet(VNode* node, unsigned int level, unsigned int position, PyObject* value) {
+  debug("doSet(): level == %i\n", level);
   if(level == 0) {
-    debug("doSet(): level == 0\n");
     VNode* theNewNode = newNode();
     copyInsert(theNewNode->items, node->items, position & BIT_MASK, value);
     incRefs((PyObject**)theNewNode->items);
     return theNewNode;
   } else {
-    debug("doSet(): level == %i\n", level);
     VNode* theNewNode = copyNode(node);
     Py_ssize_t index = (position >> level) & BIT_MASK;
 
     // Drop reference to this node since we're about to replace it
-    ((VNode*)theNewNode->items[index])->refCount--;
+    DEC_NODE_REF_COUNT((VNode*)theNewNode->items[index]);
     theNewNode->items[index] = doSet(node->items[index], level - SHIFT, position, value); 
     return theNewNode;
   }
 }
 
+
 static PyObject* internalSet(PVector *self, Py_ssize_t position, PyObject *argObj) {
   if(position < 0) {
     position += self->count;
@@ -855,7 +916,7 @@ static PyObject* internalSet(PVector *self, Py_ssize_t position, PyObject *argOb
   if((0 <= position) && (position < self->count)) {
     if(position >= TAIL_OFF(self)) {
       // Reuse the root, replace the tail
-      self->root->refCount++;
+      INC_NODE_REF_COUNT(self->root);
       PVector *new_pvec = newPvec(self->count, self->shift, self->root);
       copyInsert(new_pvec->tail->items, self->tail->items, position & BIT_MASK, argObj);
       incRefs((PyObject**)new_pvec->tail->items);
@@ -868,10 +929,11 @@ static PyObject* internalSet(PVector *self, Py_ssize_t position, PyObject *argOb
       // Free the tail and replace it with a reference to the tail of the original vector
       freeNode(new_pvec->tail);
       new_pvec->tail = self->tail;
-      self->tail->refCount++;
+      INC_NODE_REF_COUNT(self->tail);
       return (PyObject*)new_pvec;
     }
   } else if (position == self->count) {
+    // TODO Remove this case?
     return PVector_append(self, argObj);
   } else {
     PyErr_SetString(PyExc_IndexError, "Index out of range");
@@ -916,6 +978,8 @@ static PyObject* PVector_set_in(PVector *self, PyObject *args) {
 	return NULL;
       }
 
+      // This is slightly scary calling out to Python code. Could
+      // there be concurrency issues if the GIL is released?
       PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
       PyObject* newItem = PyObject_CallMethodObjArgs(currentItem, set_in_fn_name, newSequence, value, NULL);
       Py_DECREF(currentItem);
@@ -946,6 +1010,27 @@ static PyObject* PVector_set(PVector *self, PyObject *args) {
   return internalSet(self, position, argObj);
 }
 
+static PyObject* PVector_mset(PVector *self, PyObject *args) {
+  Py_ssize_t size = PyTuple_Size(args);
+  if(size % 2) {
+    PyErr_SetString(PyExc_TypeError, "mset expected an even number of arguments");
+    return NULL;
+  }
+
+  PVectorEvolver* evolver = (PVectorEvolver*)PVector_evolver(self);
+  Py_ssize_t i;
+  for(i=0; i<size; i+=2) {
+    if(PVectorEvolver_set_item(evolver, PyTuple_GetItem(args, i), PyTuple_GetItem(args, i + 1)) < 0) {
+      Py_DECREF(evolver);
+      return NULL;
+    }
+  }
+
+  PyObject* vector = PVectorEvolver_pvector(evolver);
+  Py_DECREF(evolver);
+  return vector;
+}
+
 
 static PyMethodDef PyrsistentMethods[] = {
   {"pvector", pyrsistent_pvec, METH_VARARGS, "Factory method for persistent vectors"},
@@ -974,8 +1059,9 @@ PyObject* moduleinit(void) {
   PVectorType.tp_init = NULL;
   PVectorType.tp_new = NULL;
 
-  if (PyType_Ready(&PVectorType) < 0)
+  if (PyType_Ready(&PVectorType) < 0) {
     return NULL;
+  }
 
   // Register with Sequence to appear as a proper sequence to the outside world
   PyObject* c = PyObject_GetAttrString(PyImport_ImportModule("collections"), "Sequence");
@@ -1050,7 +1136,7 @@ static PyMethodDef PVectorIter_methods[] = {
     {NULL,              NULL}           /* sentinel */
 };
 
-PyTypeObject PVectorIterType = {
+static PyTypeObject PVectorIterType = {
     PyVarObject_HEAD_INIT(&PyType_Type, 0)
     "pvector_iterator",                         /* tp_name */
     sizeof(PVectorIter),                        /* tp_basicsize */
@@ -1083,10 +1169,12 @@ PyTypeObject PVectorIterType = {
     0,                                          /* tp_members */
 };
 
-static PyObject * PVectorIter_iter(PyObject *seq) {
+static PyObject *PVectorIter_iter(PyObject *seq) {
     PVectorIter *it = PyObject_GC_New(PVectorIter, &PVectorIterType);
-    if (it == NULL)
+    if (it == NULL) {
         return NULL;
+    }
+
     it->it_index = 0;
     Py_INCREF(seq);
     it->it_seq = (PVector *)seq;
@@ -1123,3 +1211,290 @@ static PyObject *PVectorIter_next(PVectorIter *it) {
     it->it_seq = NULL;
     return NULL;
 }
+
+
+/*********************** PVector Evolver **************************/
+
+/* 
+Evolver to make multi updates easier to work with and more efficient.
+*/
+
+static void PVectorEvolver_dealloc(PVectorEvolver *);
+static PyObject *PVectorEvolver_append(PVectorEvolver *, PyObject *);
+static PyObject *PVectorEvolver_extend(PVectorEvolver *, PyObject *);
+static PyObject *PVectorEvolver_subscript(PVectorEvolver *, PyObject *);
+static PyObject *PVectorEvolver_pvector(PVectorEvolver *);
+static Py_ssize_t PVectorEvolver_len(PVectorEvolver *);
+static PyObject *PVectorEvolver_is_dirty(PVectorEvolver *);
+static int PVectorEvolver_traverse(PVectorEvolver *self, visitproc visit, void *arg);
+
+static PyMappingMethods PVectorEvolver_mapping_methods = {
+  (lenfunc)PVectorEvolver_len,
+  (binaryfunc)PVectorEvolver_subscript,
+  (objobjargproc)PVectorEvolver_set_item,
+};
+
+
+static PyMethodDef PVectorEvolver_methods[] = {
+	{"append",      (PyCFunction)PVectorEvolver_append, METH_O,       "Appends an element"},
+	{"extend",      (PyCFunction)PVectorEvolver_extend, METH_O|METH_COEXIST, "Extend"},
+        {"pvector",     (PyCFunction)PVectorEvolver_pvector, METH_NOARGS, "Create PVector from evolver"},
+        {"is_dirty",    (PyCFunction)PVectorEvolver_is_dirty, METH_NOARGS, "Check if evolver contains modifications"},
+        {NULL,              NULL}           /* sentinel */
+};
+
+static PyTypeObject PVectorEvolverType = {
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
+    "pvector_evolver",                          /* tp_name */
+    sizeof(PVectorEvolver),                     /* tp_basicsize */
+    0,                                          /* tp_itemsize */
+    /* methods */
+    (destructor)PVectorEvolver_dealloc,         /* tp_dealloc */
+    0,                                          /* tp_print */
+    0,                                          /* tp_getattr */
+    0,                                          /* tp_setattr */
+    0,                                          /* tp_compare */
+    0,                                          /* tp_repr */
+    0,                                          /* tp_as_number */
+    0,                                          /* tp_as_sequence */
+    &PVectorEvolver_mapping_methods,             /* tp_as_mapping */
+    0,                                          /* tp_hash */
+    0,                                          /* tp_call */
+    0,                                          /* tp_str */
+    PyObject_GenericGetAttr,                    /* tp_getattro */
+    0,                                          /* tp_setattro */
+    0,                                          /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,    /* tp_flags */
+    0,                                          /* tp_doc */
+    (traverseproc)PVectorEvolver_traverse,      /* tp_traverse       */
+    0,                                          /* tp_clear */
+    0,                                          /* tp_richcompare */
+    0,                                          /* tp_weaklistoffset */
+    0,                                          /* tp_iter */
+    0,                                          /* tp_iternext */
+    PVectorEvolver_methods,                     /* tp_methods */
+    0,                                          /* tp_members */
+};
+
+
+// Indicate that a node is "dirty" (has been updated by the evolver)
+// by setting the MSB of the refCount. This will be cleared when
+// creating a pvector from the evolver (cleaning it).
+#define DIRTY_BIT 0x80000000
+#define REF_COUNT_MASK (~DIRTY_BIT)
+#define IS_DIRTY(node) ((node)->refCount & DIRTY_BIT)
+#define SET_DIRTY(node) ((node)->refCount |= DIRTY_BIT)
+#define CLEAR_DIRTY(node) ((node)->refCount &= REF_COUNT_MASK)
+
+
+static void cleanNodeRecursively(VNode *node, int level) {
+  debug("Freezing recursively node=%p, level=%u\n", node, level);
+
+  int i;
+  CLEAR_DIRTY(node);
+  SET_NODE_REF_COUNT(node, 1);
+  if(level > 0) {
+    for(i = 0; i < BRANCH_FACTOR; i++) {
+      VNode *nextNode = (VNode*)node->items[i];
+      if((nextNode != NULL) && IS_DIRTY(nextNode)) {
+        cleanNodeRecursively(nextNode, level - SHIFT);
+      }
+    }
+  }
+}
+
+static void cleanVector(PVector *vector) {
+  // Freezing the vector means that all dirty indications are cleared
+  // and that the nodes that were dirty get a ref count of 1 since
+  // they are brand new. Once frozen the vector can be released into
+  // the wild.
+  if(IS_DIRTY(vector->tail)) {
+    cleanNodeRecursively(vector->tail, 0);
+  } else {
+    INC_NODE_REF_COUNT(vector->tail);
+  }
+
+  if(IS_DIRTY(vector->root)) {
+    cleanNodeRecursively(vector->root, vector->shift);
+  } else {
+    INC_NODE_REF_COUNT(vector->root);
+  }
+}
+
+static void PVectorEvolver_dealloc(PVectorEvolver *self) {
+  PyObject_GC_UnTrack(self);
+  Py_TRASHCAN_SAFE_BEGIN(self);
+
+  if(self->originalVector != self->newVector) {
+    cleanVector(self->newVector);
+    Py_DECREF(self->newVector);
+  }
+
+  Py_DECREF(self->originalVector);
+  Py_DECREF(self->appendList);
+
+  PyObject_GC_Del(self);
+  Py_TRASHCAN_SAFE_END(self);
+}
+
+static PyObject *PVectorEvolver_append(PVectorEvolver *self, PyObject *args) {
+  if (PyList_Append(self->appendList, args) == 0) {
+    Py_RETURN_NONE;
+  }
+
+  return NULL;
+}
+
+static PyObject *PVectorEvolver_extend(PVectorEvolver *self, PyObject *args) {
+  return _PyList_Extend((PyListObject *)self->appendList, args);
+}
+
+static PyObject *PVectorEvolver_subscript(PVectorEvolver *self, PyObject *item) {
+  if (PyIndex_Check(item)) {
+    Py_ssize_t position = PyNumber_AsSsize_t(item, PyExc_IndexError);
+    if (position == -1 && PyErr_Occurred()) {
+      return NULL;
+    }
+
+    if (position < 0) {
+      position += self->newVector->count + PyList_GET_SIZE(self->appendList);
+    }
+
+    if(0 <= position && position < self->newVector->count) {
+      PyObject *result = _get_item(self->newVector, position);
+      Py_XINCREF(result);
+      return result;
+    } else if (0 <= position && position < (self->newVector->count + PyList_GET_SIZE(self->appendList))) {
+      PyObject *result = PyList_GetItem(self->appendList, position - self->newVector->count);
+      Py_INCREF(result);
+      return result;
+    } else {
+      PyErr_SetString(PyExc_IndexError, "Index out of range");
+    }
+  } else {
+    PyErr_Format(PyExc_TypeError, "Indices must be integers, not %.200s", item->ob_type->tp_name);
+  }
+
+  return NULL;
+}
+
+static VNode* doSetWithDirty(VNode* node, unsigned int level, unsigned int position, PyObject* value) {
+  VNode* resultNode;
+  debug("doSetWithDirty(): level == %i\n", level);
+  if(level == 0) {
+    if(!IS_DIRTY(node)) {
+      resultNode = allocNode();
+      copyInsert(resultNode->items, node->items, position & BIT_MASK, value);
+      incRefs((PyObject**)resultNode->items);
+      SET_DIRTY(resultNode);
+    } else {
+      resultNode = node;
+      Py_INCREF(value);
+      Py_DECREF(resultNode->items[position & BIT_MASK]);
+      resultNode->items[position & BIT_MASK] = value;
+    }
+  } else {
+    if(!IS_DIRTY(node)) {
+      resultNode = copyNode(node);
+      SET_DIRTY(resultNode);
+    } else {
+      resultNode = node;
+    }    
+
+    Py_ssize_t index = (position >> level) & BIT_MASK;
+    VNode* oldNode = (VNode*)resultNode->items[index];
+    resultNode->items[index] = doSetWithDirty(resultNode->items[index], level - SHIFT, position, value);
+
+    if(resultNode->items[index] != oldNode) {
+      // Node replaced, drop references to old node
+      DEC_NODE_REF_COUNT(oldNode);
+    }
+  }
+
+  return resultNode;
+}
+
+static int PVectorEvolver_set_item(PVectorEvolver *self, PyObject* item, PyObject* value) {
+  if (PyIndex_Check(item)) {
+    Py_ssize_t position = PyNumber_AsSsize_t(item, PyExc_IndexError);
+    if (position == -1 && PyErr_Occurred()) {
+      return -1;
+    }
+         
+    if (position < 0) {
+      position += self->newVector->count + PyList_GET_SIZE(self->appendList);
+    }
+
+    if((0 <= position) && (position < self->newVector->count)) {
+      if(self->originalVector == self->newVector) {
+        // Create new vector since we're about to modify the original
+        self->newVector = rawCopyPVector(self->originalVector);
+      }
+
+      if(position < TAIL_OFF(self->newVector)) {
+        self->newVector->root = doSetWithDirty(self->newVector->root, self->newVector->shift, position, value);
+      } else {
+        self->newVector->tail = doSetWithDirty(self->newVector->tail, 0, position, value);
+      }
+      
+      return 0;
+    } else if((0 <= position) && (position < (self->newVector->count + PyList_GET_SIZE(self->appendList)))) {
+      int result = PyList_SetItem(self->appendList, position - self->newVector->count, value); 
+      if(result == 0) {
+        Py_INCREF(value);
+      }
+      return result;
+    } else if((0 <= position) && (position < (self->newVector->count + PyList_GET_SIZE(self->appendList) + 1))) {
+      return PyList_Append(self->appendList, value); 
+    } else {
+      PyErr_SetString(PyExc_IndexError, "Index out of range");
+    }
+  } else {
+    PyErr_Format(PyExc_TypeError, "Indices must be integers, not %.200s", item->ob_type->tp_name);
+  }
+  return -1;
+}
+
+static PyObject *PVectorEvolver_pvector(PVectorEvolver *self) {
+  PVector *resultVector;
+  if(self->newVector == self->originalVector) {
+    resultVector = self->newVector;
+  } else {
+    cleanVector(self->newVector);
+    resultVector = self->newVector;
+    Py_DECREF(self->originalVector);
+  }
+
+  if(PyList_GET_SIZE(self->appendList)) {
+    PVector *oldVector = resultVector;
+    resultVector = (PVector*)PVector_extend(resultVector, self->appendList);
+    Py_DECREF(oldVector);
+    Py_DECREF(self->appendList);
+    self->appendList = NULL;
+  }
+
+  initializeEvolver(self, resultVector, self->appendList);
+  Py_INCREF(resultVector);  
+  return (PyObject*)resultVector;
+}
+
+static Py_ssize_t PVectorEvolver_len(PVectorEvolver *self) {
+  return self->newVector->count + PyList_GET_SIZE(self->appendList);
+}
+
+static PyObject* PVectorEvolver_is_dirty(PVectorEvolver *self) {
+  if((self->newVector != self->originalVector) || (PyList_GET_SIZE(self->appendList) > 0)) {
+    Py_INCREF(Py_True);
+    return Py_True;
+  }
+
+  Py_INCREF(Py_False);
+  return Py_False;
+}
+
+static int PVectorEvolver_traverse(PVectorEvolver *self, visitproc visit, void *arg) {
+  Py_VISIT(self->newVector);
+  Py_VISIT(self->originalVector);
+  Py_VISIT(self->appendList);
+  return 0;
+}
