@@ -1,7 +1,9 @@
+# -*- coding: utf-8 -*-
 from collections import (Sequence, Mapping, Set, Hashable, Container, Iterable, Sized)
-from itertools import chain, islice
 from functools import wraps, reduce
+from itertools import chain, islice
 from numbers import Integral
+import sys
 
 import six
 
@@ -28,11 +30,11 @@ class PVector(object):
     Persistent vector implementation. Meant as a replacement for the cases where you would normally
     use a Python list.
 
-    Do not instantiate directly, instead use the factory functions :py:func:`v` and :func:`pvector` to
+    Do not instantiate directly, instead use the factory functions :py:func:`v` and :py:func:`pvector` to
     create an instance.
 
     Heavily influenced by the persistent vector available in Clojure. Initially this was more or
-    less just a port of the Java code for the Clojure data structures. It has since been modified and to
+    less just a port of the Java code for the Clojure vector. It has since been modified and to
     some extent optimized for usage in Python.
 
     The vector is organized as a trie, any mutating method will return a new vector that contains the changes. No
@@ -45,6 +47,8 @@ class PVector(object):
 
     The PVector implements the Sequence protocol and is Hashable.
 
+    Inserts are amortized O(1). Random access is log32(n) where n is the size of the vector.
+
     The following are examples of some common operations on persistent vectors:
 
     >>> p = v(1, 2, 3)
@@ -64,12 +68,12 @@ class PVector(object):
     """
     __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset')
 
-    def __new__(cls, c, s, r, t):
+    def __new__(cls, count, shift, root, tail):
         self = super(PVector, cls).__new__(cls)
-        self._count = c
-        self._shift = s
-        self._root = r
-        self._tail = t
+        self._count = count
+        self._shift = shift
+        self._root = root
+        self._tail = tail
 
         # Derived attribute stored for performance
         self._tail_offset = self._count - len(self._tail)
@@ -105,7 +109,7 @@ class PVector(object):
         if index < 0:
             index += self._count
 
-        return self._node_for(index)[index & BIT_MASK]
+        return PVector._node_for(self, index)[index & BIT_MASK]
 
     def __add__(self, other):
         """
@@ -117,7 +121,7 @@ class PVector(object):
         return self.extend(other)
 
     def __repr__(self):
-        return 'pvector({})'.format(str(self._tolist()))
+        return 'pvector({0})'.format(str(self._tolist()))
 
     __str__ = __repr__
 
@@ -198,11 +202,163 @@ class PVector(object):
         # Taking the easy way out again...
         return hash(self._totuple())
 
+    class _Evolver(object):
+        __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset', '_dirty_nodes', '_extra_tail', '_cached_leafs')
+
+        def __init__(self, v):
+            self._reset(v)
+
+        def __getitem__(self, index):
+            if not isinstance(index, Integral):
+                raise TypeError("'%s' object cannot be interpreted as an index" % type(index).__name__)
+
+            if index < 0:
+                index += self._count + len(self._extra_tail)
+
+            if self._count <= index < self._count + len(self._extra_tail):
+                return self._extra_tail[index - self._count]
+
+            return PVector._node_for(self, index)[index & BIT_MASK]
+
+        def _reset(self, v):
+            self._count = v._count
+            self._shift = v._shift
+            self._root = v._root
+            self._tail = v._tail
+            self._tail_offset = v._tail_offset
+            self._dirty_nodes = {}
+            self._cached_leafs = {}
+            self._extra_tail = []
+
+        def append(self, element):
+            self._extra_tail.append(element)
+
+        def extend(self, iterable):
+            self._extra_tail.extend(iterable)
+
+        def __setitem__(self, index, val):
+            if not isinstance(index, Integral):
+                raise TypeError("'%s' object cannot be interpreted as an index" % type(index).__name__)
+
+            if index < 0:
+                index += self._count + len(self._extra_tail)
+
+            if 0 <= index < self._count:
+                node = self._cached_leafs.get(index >> SHIFT)
+                if node:
+                    node[index & BIT_MASK] = val
+                elif index >= self._tail_offset:
+                    if id(self._tail) not in self._dirty_nodes:
+                        self._tail = list(self._tail)
+                        self._dirty_nodes[id(self._tail)] = True
+                        self._cached_leafs[index >> SHIFT] = self._tail
+                    self._tail[index & BIT_MASK] = val
+                else:
+                    self._root = self._do_set(self._shift, self._root, index, val)
+            elif self._count <= index < self._count + len(self._extra_tail):
+                self._extra_tail[index - self._count] = val
+            elif index == self._count + len(self._extra_tail):
+                self._extra_tail.append(val)
+            else:
+                raise IndexError()
+
+        def _do_set(self, level, node, i, val):
+            if id(node) in self._dirty_nodes:
+                ret = node
+            else:
+                ret = list(node)
+                self._dirty_nodes[id(ret)] = True
+
+            if level == 0:
+                ret[i & BIT_MASK] = val
+                self._cached_leafs[i >> SHIFT] = ret
+            else:
+                sub_index = (i >> level) & BIT_MASK  # >>>
+                ret[sub_index] = self._do_set(level - SHIFT, node[sub_index], i, val)
+
+            return ret
+
+        def pvector(self):
+            v = PVector(self._count, self._shift, self._root, self._tail).extend(self._extra_tail)
+            self._reset(v)
+            return v
+
+        def __len__(self):
+            return self._count + len(self._extra_tail)
+
+        def is_dirty(self):
+            return self._dirty_nodes or self._extra_tail
+
+    def evolver(self):
+        """
+        Create a new evolver for this pvector. The evolver acts as a mutable view of the vector
+        with "transaction like" semantics. No part of the underlying vector i updated, it is still
+        fully immutable. Furthermore multiple evolvers created from the same pvector do not
+        interfere with each other.
+
+        You may want to use an evolver instead of working directly with the pvector in the
+        following cases:
+
+        * Multiple updates are done to the same vector and the intermediate results are of no
+          interest. In this case using an evolver may be a more efficient and easier to work with.
+        * You need to pass a vector into a legacy function or a function that you have no control
+          over which performs in place mutations of lists. In this case pass an evolver instance
+          instead and then create a new pvector from the evolver once the function returns.
+
+        The following example illustrates a typical workflow when working with evolvers. It also
+        displays most of the API (which i kept small by design, you should not be tempted to
+        use evolvers in excess ;-)).
+
+        Create the evolver and perform various mutating updates to it:
+        >>> v1 = v(1, 2, 3, 4, 5)
+        >>> e = v1.evolver()
+        >>> e[1] = 22
+        >>> e.append(6)
+        >>> e.extend([7, 8, 9])
+        >>> e[8] += 1
+        >>> len(e)
+        9
+
+        The underlying pvector remains the same:
+        >>> v1
+        pvector([1, 2, 3, 4, 5])
+
+        The changes are kept in the evolver. An updated pvector can be created using the
+        pvector() function on the evolver.
+        >>> v2 = e.pvector()
+        >>> v2
+        pvector([1, 22, 3, 4, 5, 6, 7, 8, 10])
+
+        The new pvector will share data with the original pvector in the same way that would have
+        been done if only using operations on the pvector.
+        """
+        return PVector._Evolver(self)
+
+    def mset(self, *args):
+        """
+        Return a new vector with elements in specified positions replaced by values (multi set).
+
+        Elements on even positions in the argument list are interpreted as indexes while
+        elements on odd positions are considered values.
+
+        >>> v1 = v(1, 2, 3)
+        >>> v1.mset(0, 11, 2, 33)
+        pvector([11, 2, 33])
+        """
+        if len(args) % 2:
+            raise TypeError("mset expected an even number of arguments")
+
+        evolver = self.evolver()
+        for i in range(0, len(args), 2):
+            evolver[args[i]] = args[i+1]
+
+        return evolver.pvector()
+
     def set(self, i, val):
         """
-        Return a new vector with element at position i replaced with val. The first vector remains unchanged.
+        Return a new vector with element at position i replaced with val. The original vector remains unchanged.
 
-        Associng a value one step beyond the end of the vector is equal to appending. Associng beyond that will
+        Setting a value one step beyond the end of the vector is equal to appending. Setting beyond that will
         result in an IndexError.
 
         >>> v1 = v(1, 2, 3)
@@ -213,6 +369,10 @@ class PVector(object):
         >>> v1.set(-1, 4)
         pvector([1, 2, 4])
         """
+        # This method could be implemented by a call to mset() but doing so would cause
+        # a ~5 X performance penalty on PyPy (considered the primary platform for this implementation
+        #  of PVector) so we're keeping this implementation for now.
+
         if not isinstance(i, Integral):
             raise TypeError("'%s' object cannot be interpreted as an index" % type(i).__name__)
 
@@ -242,13 +402,14 @@ class PVector(object):
 
         return ret
 
-    def _node_for(self, i):
-        if 0 <= i < self._count:
-            if i >= self._tail_offset:
-                return self._tail
+    @staticmethod
+    def _node_for(pvector_like, i):
+        if 0 <= i < pvector_like._count:
+            if i >= pvector_like._tail_offset:
+                return pvector_like._tail
 
-            node = self._root
-            for level in range(self._shift, 0, -SHIFT):
+            node = pvector_like._root
+            for level in range(pvector_like._shift, 0, -SHIFT):
                 node = node[(i >> level) & BIT_MASK]  # >>>
 
             return node
@@ -402,7 +563,6 @@ class PVector(object):
         return _pvector, (self._tolist(),)
 
 
-
 Sequence.register(PVector)
 Hashable.register(PVector)
 
@@ -431,7 +591,7 @@ except ImportError:
 
 def v(*elements):
     """
-    Create a new persistent vector containing all paramenters to this function.
+    Create a new persistent vector containing all parameters to this function.
 
     >>> v1 = v(1, 2, 3)
     >>> v1
@@ -460,6 +620,8 @@ class PMap(object):
 
     PMap implements the Mapping protocol and is Hashable.
 
+    Random access and insert is log32(n) where n is the size of the map.
+
     The following are examples of some common operations on persistent maps
 
     >>> m1 = m(a=1, b=3)
@@ -480,13 +642,15 @@ class PMap(object):
         self._size = size
         self._buckets = buckets
 
-    def _get_bucket(self, key):
-        index = hash(key) % len(self._buckets)
-        bucket = self._buckets[index]
+    @staticmethod
+    def _get_bucket(buckets, key):
+        index = hash(key) % len(buckets)
+        bucket = buckets[index]
         return index, bucket
 
-    def __getitem__(self, key):
-        _, bucket = self._get_bucket(key)
+    @staticmethod
+    def _getitem(buckets, key):
+        _, bucket = PMap._get_bucket(buckets, key)
         if bucket:
             for k, v in bucket:
                 if k == key:
@@ -494,8 +658,12 @@ class PMap(object):
 
         raise KeyError
 
-    def __contains__(self, key):
-        _, bucket = self._get_bucket(key)
+    def __getitem__(self, key):
+        return PMap._getitem(self._buckets, key)
+
+    @staticmethod
+    def _contains(buckets, key):
+        _, bucket = PMap._get_bucket(buckets, key)
         if bucket:
             for k, _ in bucket:
                 if k == key:
@@ -505,6 +673,9 @@ class PMap(object):
 
         return False
 
+    def __contains__(self, key):
+        return self._contains(self._buckets, key)
+
     get = Mapping.get
 
     def __iter__(self):
@@ -543,7 +714,7 @@ class PMap(object):
         return self._size
 
     def __repr__(self):
-        return 'pmap({})'.format(str(dict(self)))
+        return 'pmap({0})'.format(str(dict(self)))
 
     __eq__ = Mapping.__eq__
     __ne__ = Mapping.__ne__
@@ -567,97 +738,71 @@ class PMap(object):
         >>> m3
         pmap({'a': 1, 'c': 4, 'b': 2})
         """
-        kv = (key, val)
-        index, bucket = self._get_bucket(key)
-        if bucket:
-            for k, v in bucket:
-                if k == key:
-                    if v is val:
-                        return self
-                    else:
-                        new_bucket = [(k2, v2) if k2 != k else (k2, val) for k2, v2 in bucket]
-                        return PMap(self._size, self._buckets.set(index, new_bucket))
-
-            if len(self._buckets) < 0.67 * self._size:
-                return PMap(self._size, self._reallocate(2 * len(self._buckets))).set(key, val)
-            else:
-                new_bucket = [kv]
-                new_bucket.extend(bucket)
-                new_buckets = self._buckets.set(index, new_bucket)
-
-            return PMap(self._size + 1, new_buckets)
-
-        # Skip reallocation check if there was no conflict
-        return PMap(self._size + 1, self._buckets.set(index, [kv]))
+        evolver = self.evolver()
+        evolver[key] = val
+        return evolver.pmap()
 
     def remove(self, key):
+        """
+        Return a new PMap without the element specified by key. Raises KeyError if the element
+        is not present.
+        >>> m1 = m(a=1, b=2)
+        >>> m1.remove('a')
+        pmap({'b': 2})
+        """
+        evolver = self.evolver()
+        del evolver[key]
+        return evolver.pmap()
+
+    def discard(self, key):
         """
         Return a new PMap without the element specified by key. Returns reference to itself
         if element is not present.
 
         >>> m1 = m(a=1, b=2)
-        >>> m1.remove('a')
+        >>> m1.discard('a')
         pmap({'b': 2})
-        >>> m1 is m1.remove('c')
+        >>> m1 is m1.discard('c')
         True
         """
+        try:
+            return self.remove(key)
+        except KeyError:
+            return self
 
-        # Should shrinking of the map ever be done if it becomes very small?
-        index, bucket = self._get_bucket(key)
-
-        if bucket:
-            new_bucket = [(k, v) for (k, v) in bucket if k != key]
-            if len(bucket) > len(new_bucket):
-                return PMap(self._size - 1, self._buckets.set(index, new_bucket if new_bucket else None))
-
-        return self
-
-    def merge(self, *maps):
+    def update(self, *maps):
         """
         Return a new PMap with the items in Mappings inserted. If the same key is present in multiple
         maps the rightmost (last) value is inserted.
 
         >>> m1 = m(a=1, b=2)
-        >>> m1.merge(m(a=2, c=3), {'a': 17, 'd': 35})
+        >>> m1.update(m(a=2, c=3), {'a': 17, 'd': 35})
         pmap({'a': 17, 'c': 3, 'b': 2, 'd': 35})
         """
-        # Optimization opportunities here
-        if not maps:
-            return self
-        elif len(maps) > 1:
-            merge_map = dict(maps[0])
-            for m in maps[1:]:
-                merge_map.update(m)
-        else:
-            merge_map = maps[0]
+        return self.update_with(lambda l, r: r, *maps)
 
-        result = self
-        for k, v in merge_map.items():
-            result = result.set(k, v)
-
-        return result
-
-    def merge_with(self, merge_fn, *maps):
+    def update_with(self, update_fn, *maps):
         """
-        Return a new PMap with the items in Mappings inserted. If the same key is present in multiple
+        Return a new PMap with the items in Mappings maps inserted. If the same key is present in multiple
         maps the values will be merged using merge_fn going from left to right.
 
         >>> from operator import add
         >>> m1 = m(a=1, b=2)
-        >>> m1.merge_with(add, m(a=2))
+        >>> m1.update_with(add, m(a=2))
         pmap({'a': 3, 'b': 2})
 
         The reverse behaviour of the regular merge. Keep the leftmost element instead of the rightmost.
+
         >>> m1 = m(a=1)
-        >>> m1.merge_with(lambda l, r: l, m(a=2), {'a':3})
+        >>> m1.update_with(lambda l, r: l, m(a=2), {'a':3})
         pmap({'a': 1})
         """
-        result = self
+        evolver = self.evolver()
         for map in maps:
             for key, value in map.items():
-                result = result.set(key, merge_fn(result[key], value) if key in result else value)
+                evolver[key] = update_fn(evolver[key], value) if key in evolver else value
 
-        return result
+        return evolver.pmap()
 
     def set_in(self, keys, val):
         """
@@ -675,24 +820,110 @@ class PMap(object):
         else:
             return self.set(keys[0], self.get(keys[0], _EMPTY_PMAP).set_in(keys[1:], val))
 
-    def _reallocate_to_list(self, new_size):
-        new_list = new_size * [None]
-        for k, v in chain.from_iterable(x for x in self._buckets if x):
-            index = hash(k) % new_size
-            if new_list[index]:
-                new_list[index].append((k, v))
-            else:
-                new_list[index] = [(k, v)]
-
-        return new_list
-
-    def _reallocate(self, new_size):
-        return pvector(self._reallocate_to_list(new_size))
-
     def __reduce__(self):
         # Pickling support
         return pmap, (dict(self),)
 
+    class _Evolver(object):
+        __slots__ = ('_buckets_evolver', '_size', '_original_pmap')
+
+        def __init__(self, original_pmap):
+            self._original_pmap = original_pmap
+            self._buckets_evolver = original_pmap._buckets.evolver()
+            self._size = original_pmap._size
+
+        def __getitem__(self, key):
+            return PMap._getitem(self._buckets_evolver, key)
+
+        def __setitem__(self, key, val):
+            if len(self._buckets_evolver) < 0.67 * self._size:
+                self._reallocate(2 * len(self._buckets_evolver))
+
+            kv = (key, val)
+            index, bucket = PMap._get_bucket(self._buckets_evolver, key)
+            if bucket:
+                for k, v in bucket:
+                    if k == key:
+                        if v is not val:
+                            new_bucket = [(k2, v2) if k2 != k else (k2, val) for k2, v2 in bucket]
+                            self._buckets_evolver[index] = new_bucket
+
+                        return
+
+                new_bucket = [kv]
+                new_bucket.extend(bucket)
+                self._buckets_evolver[index] = new_bucket
+                self._size += 1
+            else:
+                self._buckets_evolver[index] = [kv]
+                self._size += 1
+
+        def _reallocate(self, new_size):
+            new_list = new_size * [None]
+            buckets = self._buckets_evolver.pvector()
+            for k, v in chain.from_iterable(x for x in buckets if x):
+                index = hash(k) % new_size
+                if new_list[index]:
+                    new_list[index].append((k, v))
+                else:
+                    new_list[index] = [(k, v)]
+
+            self._buckets_evolver = pvector(new_list).evolver()
+
+        def is_dirty(self):
+            return self._buckets_evolver.is_dirty()
+
+        def pmap(self):
+            if self.is_dirty():
+                return PMap(self._size, self._buckets_evolver.pvector())
+
+            return self._original_pmap
+
+        def __len__(self):
+            return self._size
+
+        def __contains__(self, key):
+            return PMap._contains(self._buckets_evolver, key)
+
+        def __delitem__(self, key):
+            index, bucket = PMap._get_bucket(self._buckets_evolver, key)
+
+            if bucket:
+                new_bucket = [(k, v) for (k, v) in bucket if k != key]
+                if len(bucket) > len(new_bucket):
+                    self._buckets_evolver[index] = new_bucket if new_bucket else None
+                    self._size -= 1
+                    return
+
+            raise KeyError('{0}'.format(key))
+
+    def evolver(self):
+        """
+        Create a new evolver for this pmap. For a discussion on evolvers in general see the
+        documentation for the pvector evolver.
+
+        Create the evolver and perform various mutating updates to it:
+        >>> m1 = m(a=1, b=2)
+        >>> e = m1.evolver()
+        >>> e['c'] = 3
+        >>> len(e)
+        3
+        >>> del e['a']
+
+        The underlying pmap remains the same:
+        >>> m1
+        pmap({'a': 1, 'b': 2})
+
+        The changes are kept in the evolver. An updated pmap can be created using the
+        pmap() function on the evolver.
+        >>> m2 = e.pmap()
+        >>> m2
+        pmap({'c': 3, 'b': 2})
+
+        The new pmap will share data with the original pmap in the same way that would have
+        been done if only using operations on the pmap.
+        """
+        return PMap._Evolver(self)
 
 Mapping.register(PMap)
 Hashable.register(PMap)
@@ -731,8 +962,8 @@ def pmap(initial={}, pre_size=0):
     may have a positive performance impact in the cases where you know beforehand that a large number of elements
     will be inserted into the map eventually since it will reduce the number of reallocations required.
 
-    >>> pmap({'a': 13, 'b': '14'})
-    pmap({'a': 13, 'b': '14'})
+    >>> pmap({'a': 13, 'b': 14})
+    pmap({'a': 13, 'b': 14})
     """
     if not initial:
         return _EMPTY_PMAP
@@ -751,6 +982,10 @@ def m(**kwargs):
 
 ##################### PSet ########################
 
+
+PY2 = sys.version_info[0] < 3
+
+
 class PSet(object):
     """
     Persistent set implementation. Built on top of the persistent map. The set supports all operations
@@ -759,6 +994,8 @@ class PSet(object):
     Do not instantiate directly, instead use the factory functions :py:func:`s` or :py:func:`pset`
     to create an instance.
 
+    Random access and insert is log32(n) where n is the size of the set.
+
     Some examples:
 
     >>> s = pset([1, 2, 3, 1])
@@ -786,10 +1023,10 @@ class PSet(object):
         return len(self._map)
 
     def __repr__(self):
-        if six.PY2 or not self:
+        if PY2 or not self:
             return 'p' + str(set(self))
 
-        return 'pset([{}])'.format(str(set(self))[1:-1])
+        return 'pset([{0}])'.format(str(set(self))[1:-1])
 
     __str__ = __repr__
 
@@ -798,7 +1035,7 @@ class PSet(object):
 
     @classmethod
     def _from_iterable(cls, it, pre_size=8):
-        return PSet(pmap({k: True for k in it}, pre_size=pre_size))
+        return PSet(pmap(dict((k, True) for k in it), pre_size=pre_size))
 
     def add(self, element):
         """
@@ -832,6 +1069,59 @@ class PSet(object):
 
         return self
 
+    class _Evolver(object):
+        __slots__ = ('_original_pset', '_pmap_evolver')
+
+        def __init__(self, original_pset):
+            self._original_pset = original_pset
+            self._pmap_evolver = original_pset._map.evolver()
+
+        def add(self, element):
+            self._pmap_evolver[element] = True
+
+        def remove(self, element):
+            del self._pmap_evolver[element]
+
+        def is_dirty(self):
+            return self._pmap_evolver.is_dirty()
+
+        def pset(self):
+            if not self.is_dirty():
+                return  self._original_pset
+
+            return PSet(self._pmap_evolver.pmap())
+
+        def __len__(self):
+            return len(self._pmap_evolver)
+
+    def evolver(self):
+        """
+        Create a new evolver for this pset. For a discussion on evolvers in general see the
+        documentation for the pvector evolver.
+
+        Create the evolver and perform various mutating updates to it:
+        >>> s1 = s(1, 2, 3)
+        >>> e = s1.evolver()
+        >>> e.add(4)
+        >>> len(e)
+        4
+        >>> e.remove(1)
+
+        The underlying pset remains the same:
+        >>> s1
+        pset([1, 2, 3])
+
+        The changes are kept in the evolver. An updated pmap can be created using the
+        pset() function on the evolver.
+        >>> s2 = e.pset()
+        >>> s2
+        pset([2, 3, 4])
+
+        The new pset will share data with the original pset in the same way that would have
+        been done if only using operations on the pset.
+        """
+        return PSet._Evolver(self)
+
     # All the operations and comparisons you would expect on a set.
     #
     # This is not very beautiful. If we avoid inheriting from PSet we can use the
@@ -1006,7 +1296,7 @@ class _PBag(object):
         return elt in self._counts
 
     def __repr__(self):
-        return "pbag({})".format(list(self))
+        return "pbag({0})".format(list(self))
 
     def __eq__(self, other):
         """
@@ -1125,7 +1415,7 @@ def pclass(members='', name='PClass', verbose=False):
         frozen_members = ["'%s'" % f for f in members if f.endswith('_')]
         if frozen_members:
             return """
-        frozen_fields = fields_to_modify & {{{frozen_members}}}
+        frozen_fields = fields_to_modify & set([{frozen_members}])
         if frozen_fields:
             raise AttributeError('Cannot set frozen members %s' % ', '.join(frozen_fields))
             """.format(frozen_members=', '.join(frozen_members))
@@ -1151,8 +1441,8 @@ class {class_name}(namedtuple('PClassBase', [{quoted_members}], verbose={verbose
         {frozen_member_test}
 
         return self.__class__.__new__(self.__class__, *map(kwargs.pop, [{quoted_members}], self))
-    """.format(quoted_members=quoted_members,
-               member_set="{%s}" % quoted_members if quoted_members else 'set()',
+""".format(quoted_members=quoted_members,
+               member_set="set([%s])" % quoted_members if quoted_members else 'set()',
                frozen_member_test=frozen_member_test(),
                verbose=verbose,
                class_name=name)
@@ -1165,6 +1455,7 @@ class {class_name}(namedtuple('PClassBase', [{quoted_members}], verbose={verbose
     try:
         six.exec_(template, namespace)
     except SyntaxError as e:
+        raise e
         raise SyntaxError(e.message + ':\n' + template)
 
     return namespace[name]
@@ -1198,7 +1489,7 @@ def freeze(o):
     """
     typ = type(o)
     if typ is dict:
-        return pmap({k: freeze(v) for k, v in six.iteritems(o)})
+        return pmap(dict((k, freeze(v)) for k, v in six.iteritems(o)))
     if typ is list:
         return pvector(map(freeze, o))
     if typ is tuple:
@@ -1228,7 +1519,7 @@ def thaw(o):
     if typ is type(pvector()):
         return list(map(thaw, o))
     if typ is type(pmap()):
-        return {k: thaw(v) for k, v in o.iteritems()}
+        return dict((k, thaw(v)) for k, v in o.iteritems())
     if typ is tuple:
         return tuple(map(thaw, o))
     if typ is type(pset()):
@@ -1289,7 +1580,7 @@ class _PListBase(object):
         return sum(1 for _ in self)
 
     def __repr__(self):
-        return "plist({})".format(list(self))
+        return "plist({0})".format(list(self))
     __str__ = __repr__
 
     def cons(self, elem):
@@ -1442,7 +1733,7 @@ class _PListBase(object):
             builder.append_elem(head.first)
             head = head.rest
 
-        raise ValueError('{} not found in PList'.format(elem))
+        raise ValueError('{0} not found in PList'.format(elem))
 
 
 class _PList(_PListBase):
@@ -1626,7 +1917,7 @@ class _PDeque(object):
 
     def __repr__(self):
         return "pdeque({0}{1})".format(list(self),
-                                       ', maxlen={}'.format(self._maxlen) if self._maxlen is not None else '')
+                                       ', maxlen={0}'.format(self._maxlen) if self._maxlen is not None else '')
     __str__ = __repr__
 
     @property
@@ -1811,7 +2102,7 @@ class _PDeque(object):
                 return _PDeque(self._left_list,
                                self._right_list.reverse().remove(elem).reverse(), self._length - 1)
             except ValueError:
-                raise ValueError('{} not found in PDeque'.format(elem))
+                raise ValueError('{0} not found in PDeque'.format(elem))
 
     def reverse(self):
         """
