@@ -1,5 +1,5 @@
-import pytest
 import pickle
+import pytest
 
 @pytest.fixture(scope='session', params=['pyrsistent', 'pvectorc'])
 def pvector(request):
@@ -313,7 +313,6 @@ def test_is_hashable(pvector):
     assert isinstance(pvector(), Hashable)
 
 
-
 def test_refuses_to_hash_when_members_are_unhashable(pvector):
     v = pvector([1, 2, [1, 2]])
 
@@ -452,12 +451,274 @@ def test_count(pvector):
     assert x.count(1) == 2
     assert x.count(4) == 0
 
+
 def test_empty_truthiness(pvector):
     assert pvector([1])
     assert not pvector([])
 
+
 def test_pickling_empty_vector(pvector):
     assert pickle.loads(pickle.dumps(pvector(), -1)) == pvector()
 
+
 def test_pickling_non_empty_vector(pvector):
-    assert pickle.loads(pickle.dumps(pvector([1, 'a']), -1)) == pvector([1, 'a'])
\ No newline at end of file
+    assert pickle.loads(pickle.dumps(pvector([1, 'a']), -1)) == pvector([1, 'a'])
+
+
+def test_mset_basic_assignments(pvector):
+    v1 = pvector(range(2000))
+    v2 = v1.mset(1, -1, 505, -505, 1998, -1998)
+
+    # Original not changed
+    assert v1[1] == 1
+    assert v1[505] == 505
+    assert v1[1998] == 1998
+
+    # Other updated
+    assert v2[1] == -1
+    assert v2[505] == -505
+    assert v2[1998] == -1998
+
+
+def test_mset_odd_number_of_arguments(pvector):
+    v = pvector([0, 1])
+
+    with pytest.raises(TypeError):
+        v.mset(0, 10, 1)
+
+
+def test_mset_index_out_of_range(pvector):
+    v = pvector([0, 1])
+
+    with pytest.raises(IndexError):
+        v.mset(3, 10)
+
+
+def test_evolver_no_update(pvector):
+    # This is mostly a test against memory leaks in the C implementation
+    v = pvector(range(40))
+
+    assert v.evolver().pvector() == v
+
+
+def test_evolver_deallocate_dirty_evolver(pvector):
+    # Ref count handling in native implementation
+    v = pvector(range(3220))
+    e = v.evolver()
+    e[10] = -10
+    e[3220] = -3220
+
+
+def test_evolver_simple_update_in_tree(pvector):
+    v = pvector(range(35))
+    e = v.evolver()
+    e[10] = -10
+
+    assert e[10] == -10
+    assert e.pvector()[10] == -10
+
+
+def test_evolver_multi_level_multi_update_in_tree(pvector):
+    # This test is mostly to detect memory/ref count issues in the native implementation
+    v = pvector(range(3500))
+    e = v.evolver()
+
+    # Update differs between first and second time since the
+    # corresponding node will be marked as dirty the first time only.
+    e[10] = -10
+    e[11] = -11
+    e[10] = -1000
+
+    # Update in neighbour node
+    e[50] = -50
+    e[50] = -5000
+
+    # Update in node in other half of vector
+    e[3000] = -3000
+    e[3000] = -30000
+
+    # Before freezing
+    assert e[10] == -1000
+    assert e[11] == -11
+    assert e[50] == -5000
+    assert e[3000] == -30000
+
+    # Run a GC to provoke any potential misbehavior
+    import gc
+    gc.collect()
+
+    v2 = e.pvector()
+    assert v2[10] == -1000
+    assert v2[50] == -5000
+    assert v2[3000] == -30000
+
+    # Run a GC to provoke any potential misbehavior
+    gc.collect()
+
+    # After freezing
+    assert e[10] == -1000
+    assert e[11] == -11
+    assert e[50] == -5000
+    assert e[3000] == -30000
+
+    # Original stays the same
+    assert v[10] == 10
+    assert v[50] == 50
+    assert v[3000] == 3000
+
+
+def test_evolver_simple_update_in_tail(pvector):
+    v = pvector(range(35))
+    e = v.evolver()
+    e[33] = -33
+
+    assert e[33] == -33
+    assert e.pvector()[33] == -33
+    assert v[33] == 33
+
+
+def test_evolver_simple_update_just_outside_vector(pvector):
+    v = pvector()
+    e = v.evolver()
+    e[0] = 1
+
+    assert e[0] == 1
+    assert e.pvector()[0] == 1
+    assert len(v) == 0
+
+
+def test_evolver_append(pvector):
+    v = pvector()
+    e = v.evolver()
+    e.append(1000)
+    assert e[0] == 1000
+
+    e[0] = 2000
+    assert e[0] == 2000
+    assert list(e.pvector()) == [2000]
+    assert list(v) == []
+
+
+def test_evolver_extend(pvector):
+    v = pvector([1000])
+    e = v.evolver()
+    e.extend([2000, 3000])
+    e[2] = 20000
+
+    assert list(e.pvector()) == [1000, 2000, 20000]
+    assert list(v) == [1000]
+
+
+def test_evolver_assign_and_read_with_negative_indices(pvector):
+    v = pvector([1, 2, 3])
+    e = v.evolver()
+    e[-1] = 4
+    e.extend([11, 12, 13])
+    e[-1] = 33
+
+    assert e[-1] == 33
+    assert list(e.pvector()) == [1, 2, 4, 11, 12, 33]
+
+
+def test_evolver_non_integral_access(pvector):
+    e = pvector([1]).evolver()
+
+    with pytest.raises(TypeError):
+        x = e['foo']
+
+
+def test_evolver_non_integral_assignment(pvector):
+    e = pvector([1]).evolver()
+
+    with pytest.raises(TypeError):
+        e['foo'] = 1
+
+
+def test_evolver_out_of_bounds_access(pvector):
+    e = pvector([1]).evolver()
+
+    with pytest.raises(IndexError):
+        x = e[1]
+
+
+def test_evolver_out_of_bounds_assignment(pvector):
+    e = pvector([1]).evolver()
+
+    with pytest.raises(IndexError):
+        e[2] = 1
+
+
+def test_no_dependencies_between_evolvers_from_the_same_pvector(pvector):
+    original_list = list(range(40))
+    v = pvector(original_list)
+    e1 = v.evolver()
+    e2 = v.evolver()
+
+    e1.extend([1, 2, 3])
+    e1[2] = 20
+    e1[35] = 350
+
+    e2.extend([-1, -2, -3])
+    e2[2] = -20
+    e2[35] = -350
+
+    e1_expected = original_list + [1, 2, 3]
+    e1_expected[2] = 20
+    e1_expected[35] = 350
+    assert list(e1.pvector()) == e1_expected
+
+    e2_expected = original_list + [-1, -2, -3]
+    e2_expected[2] = -20
+    e2_expected[35] = -350
+    assert list(e2.pvector()) == e2_expected
+
+
+def test_pvectors_produced_from_the_same_evolver_do_not_interfere(pvector):
+    original_list = list(range(40))
+    v = pvector(original_list)
+    e = v.evolver()
+
+    e.extend([1, 2, 3])
+    e[2] = 20
+    e[35] = 350
+
+    v1 = e.pvector()
+    v1_expected = original_list + [1, 2, 3]
+    v1_expected[2] = 20
+    v1_expected[35] = 350
+
+    e.extend([-1, -2, -3])
+    e[3] = -30
+    e[36] = -360
+
+    v2 = e.pvector()
+    v2_expected = v1_expected + [-1, -2, -3]
+    v2_expected[3] = -30
+    v2_expected[36] = -360
+
+    assert list(v1) == v1_expected
+    assert list(v2) == v2_expected
+
+
+def test_evolver_len(pvector):
+    e = pvector([1, 2, 3]).evolver()
+    e.extend([4, 5])
+
+    assert len(e) == 5
+
+
+def test_evolver_is_dirty(pvector):
+    e = pvector([1, 2, 3]).evolver()
+    assert not e.is_dirty()
+
+    e.append(4)
+    assert e.is_dirty
+
+    e.pvector()
+    assert not e.is_dirty()
+
+    e[2] = 2000
+    assert e.is_dirty
+
+    e.pvector()
+    assert not e.is_dirty()
